#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from pyxcp.master.base import MasterBaseType
from pyxcp.master.errorhandler import wrapped
from pyxcp import types


class Master(MasterBaseType):
    # Python >= 3.5 permits nice unpacking syntax (PEP448).

    @wrapped
    def shortDownload(self, address, addressExt, data):
        length = len(data)
        addr = self.DWORD_pack(address)
        response = self.transport.request(
            types.Command.SHORT_DOWNLOAD, length, 0, addressExt, *addr, *data)
        return response

    @wrapped
    def modifyBits(self, shiftValue, andMask, xorMask):
        # A = ( (A) & ((~((dword)(((word)~MA)<<S))) )^((dword)(MX<<S)) )
        am = self.WORD_pack(andMask)
        xm = self.WORD_pack(xorMask)
        response = self.transport.request(
            types.Command.MODIFY_BITS, shiftValue, *am, *xm)
        return response

    @wrapped
    def setDaqPtr(self, daqListNumber, odtNumber, odtEntryNumber):
        self.currentDaqPtr = types.DaqPtr(daqListNumber, odtNumber, odtEntryNumber) # Needed for errorhandling.
        daqList = self.WORD_pack(daqListNumber)
        response = self.transport.request(
            types.Command.SET_DAQ_PTR, 0, *daqList, odtNumber, odtEntryNumber)
        return response

    @wrapped
    def setDaqListMode(self, mode, daqListNumber, eventChannelNumber,
                       prescaler, priority):
        dln = self.WORD_pack(daqListNumber)
        ecn = self.WORD_pack(eventChannelNumber)
        response = self.transport.request(
            types.Command.SET_DAQ_LIST_MODE,
            mode, *dln, *ecn, prescaler, priority)
        return response

    @wrapped
    def allocOdt(self, daqListNumber, odtCount):
        dln = self.WORD_pack(daqListNumber)
        response = self.transport.request(
            types.Command.ALLOC_ODT, 0, *dln, odtCount)
        return response

    @wrapped
    def allocOdtEntry(self, daqListNumber, odtNumber, odtEntriesCount):
        dln = self.WORD_pack(daqListNumber)
        response = self.transport.request(
            types.Command.ALLOC_ODT_ENTRY, 0, *dln, odtNumber, odtEntriesCount)
        return response




from sys import version_info

from pyxcp.utils import (hexDump, slicer, flatten,
    getPythonVersion, PYTHON_VERSION)

def test_hexdump(capsys):
    print(hexDump(range(16)), end = "")
    captured = capsys.readouterr()
    assert captured.out == "[00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f]"

def test_slicer1():
    res = slicer([1,2,3,4,5,6,7,8], 4)
    assert res == [[1, 2, 3, 4], [5, 6, 7, 8]]

def test_slicer2():
    res = slicer(['10', '20', '30', '40', '50', '60', '70', '80'], 4, tuple)
    assert res == [('10', '20', '30', '40'), ('50', '60', '70', '80')]

def test_flatten1():
    res = flatten([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert res == [1, 2, 3, 4, 5, 6, 7, 8]

def test_version():
    assert getPythonVersion() == version_info
    assert PYTHON_VERSION == version_info
    assert getPythonVersion() == PYTHON_VERSION




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parse (transport-layer specific) command line parameters
and create a XCP master instance.

.. note::
    There is currently no interface for further customization.

"""

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


import argparse


from pyxcp.config import readConfiguration
from pyxcp.master import Master
from pyxcp.transport.can import (try_to_install_system_supplied_drivers, registered_drivers)

try_to_install_system_supplied_drivers()

CAN_DRIVERS = registered_drivers()


CMD_LINE_ARGUMENTS = {
    "can": ("CAN_DRIVER", "LOGLEVEL"),
    "eth": ("HOST", "PORT", "PROTOCOL", "IPV6", "LOGLEVEL"),
    "sxi": ("PORT", "BAUDRATE", "BYTESIZE", "PARITY", "STOPBITS", "LOGLEVEL"),
}

def merge_parameters(transport, params_from_cfg_file, params_from_cmd_line):
    """Merge parameters from config-file and command-line.
    The latter have precedence.

    Parameters
    ----------
    params_from_cfg_file: dict

    params_from_cmd_line: dict

    Returns
    -------
    dict
    """
    args = CMD_LINE_ARGUMENTS.get(transport)
    params_from_cmd_line = {k.upper(): v for k, v in params_from_cmd_line.items()}
    result = {}
    for arg in args:
        cvalue = params_from_cfg_file.get(arg)
        if cvalue:
            result[arg] = cvalue
        pvalue = params_from_cmd_line.get(arg)
        if pvalue:
            result[arg] = pvalue
    return result



def makeNonNullValuesDict(**params):
    """Only add items with non-None values.
    """
    return {k: v for k, v in params.items() if not v is None}


class ArgumentParser:
    """

    """

    def __init__(self, *args, **kws):
        kws.update(formatter_class = argparse.RawDescriptionHelpFormatter)
        self.parser = argparse.ArgumentParser(*args, **kws)
        subparsers = self.parser.add_subparsers(dest = "transport")
        subparsers.help = "Transport layer"
        self.parser.add_argument('-c', '--config-file', type=argparse.FileType('r'), dest = "conf",
            help = 'File to read (extended) parameters from.')
        self.parser.add_argument('-l', '--loglevel', choices = ["ERROR", "WARN", "INFO", "DEBUG"])
        self.parser.epilog = "To get specific help on transport layers\nuse <layer> -h, e.g. {} eth -h".format(self.parser.prog)

        eth = subparsers.add_parser("eth", description = "XCPonEth specific options:")
        eth.set_defaults(eth = True)
        sxi = subparsers.add_parser("sxi", description = "XCPonSxI specific options:")
        sxi.set_defaults(sxi = True)
        # TODO: conditionally add CAN options (only if at least one driver available")
        can = subparsers.add_parser("can", description = "XCPonCAN specific options:")
        can.set_defaults(can = True)

        can.add_argument('-d', '--driver', choices = CAN_DRIVERS.keys(), dest = "can_driver")

        eth.add_argument('-p', '--port', type = int, metavar = "port")
        proto = eth.add_mutually_exclusive_group()
        proto.add_argument('-t', '--tcp', default = True, const = True, metavar = "tcp", action = "store_const")
        proto.add_argument('-u', '--udp', default = False, const = True, metavar = "udp", action = "store_const")

        eth.add_argument('-6', '--ipv6', const = True, metavar = "ipv6", action = "store_const")
        eth.add_argument('-H', '--host', help = "Host name or IP.")

        sxi.add_argument('-p', '--port', help = "Name or number of your serial interface.")
        sxi.add_argument('-b', '--baudrate', type = int)
        sxi.add_argument('--bytesize', type = int)
        sxi.add_argument('--parity', choices = ['N', 'E', 'O'])
        sxi.add_argument('--stopbits', type = int, choices = [1, 2])
        self._args = []


    def run(self):
        """

        """
        self._args = self.parser.parse_args()
        args = self.args
        config = readConfiguration(args.conf)
        transport = args.transport
        if not transport:
            print("missing argument transport: choose from {}".format(['can', 'eth', 'sxi']))
            exit(1)

        if transport == "eth":
            params = makeNonNullValuesDict(
                host = args.host,
                port = args.port,
                protocol = "UDP" if args.udp else "TCP",
                ipv6 = args.ipv6,
                loglevel = args.loglevel)
        elif transport == "sxi":
            params = makeNonNullValuesDict(
                port = args.port,
                baudrate = args.baudrate,
                bytesize = args.bytesize,
                parity = args.parity,
                stopbits = args.stopbits,
                loglevel = args.loglevel)
        elif transport == "can":
            if not args.can_driver in CAN_DRIVERS:
                print("missing argument CAN driver (-d <driver>): choose from {}".format([x for x in CAN_DRIVERS.keys()]))
                exit(1)
            params = dict(
                loglevel = args.loglevel,
                can_driver = args.can_driver
            )

        params = merge_parameters(transport, config, params)
        config.update(params)
        return Master(transport, config = config)

    @property
    def args(self):
        return self._args




from collections import OrderedDict
from io import StringIO

from pyxcp.config import readConfiguration

JSON = """{
    "PORT": "COM10",
    "BITRATE": 38400,
    "BYTESIZE": 8,
    "PARITY": "N",
    "STOPBITS": 1,
    "CREATE_DAQ_TIMESTAMPS": false
}"""

TOML = """PORT = "COM10"
BITRATE = 38400
PARITY = "N"
BYTESIZE = 8
STOPBITS = 1
CREATE_DAQ_TIMESTAMPS = false
"""

CONF_JSON = StringIO(JSON)
CONF_JSON.name = "hello.json"

CONF_TOML = StringIO(TOML)
CONF_TOML.name = "hello.toml"


def test_read_empty_config():
    assert readConfiguration(None) == {}
    assert readConfiguration({}) == {}


def test_read_json_config():
    assert readConfiguration(CONF_JSON) == {'BITRATE': 38400,
                                            'BYTESIZE': 8,
                                            'CREATE_DAQ_TIMESTAMPS': False,
                                            'PARITY': 'N',
                                            'PORT': 'COM10',
                                            'STOPBITS': 1}


def test_read_toml_config():
    assert readConfiguration(CONF_TOML) == {'BITRATE': 38400,
                                            'BYTESIZE': 8,
                                            'CREATE_DAQ_TIMESTAMPS': False,
                                            'PARITY': 'N',
                                            'PORT': 'COM10',
                                            'STOPBITS': 1}


def test_read_dict():
    assert readConfiguration({'A': 1, 'B': 2, 'C': 3}) == \
        {'A': 1, 'B': 2, 'C': 3}
    assert readConfiguration(OrderedDict({'A': 1, 'B': 2, 'C': 3})) == \
        {'A': 1, 'B': 2, 'C': 3}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2018 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import ctypes
import mmap
import os
import sys
#import subprocess
import threading

##
##try:
##    import win32api
##    import win32process
##    import win32con
##except ImportError:
##    WINAPI = False
##else:
##    WINAPI = True
##


def hexDump(arr):
    return "[{}]".format(' '.join(["{:02x}".format(x) for x in arr]))


def slicer(iterable, sliceLength, converter=None):
    if converter is None:
        converter = type(iterable)
    length = len(iterable)
    return [
        converter((iterable[item:item + sliceLength]))
        for item in range(0, length, sliceLength)]


def flatten(*args):
    result = []
    for arg in list(args):
        if hasattr(arg, '__iter__'):
            result.extend(flatten(*arg))
        else:
            result.append(arg)
    return result


##
##def intToArray(value):
##    result = []
##    while value:
##        result.append(value & 0xff)
##        value >>= 8
##    if result:
##        return list(reversed(result))
##    else:
##        return [0]
##
##
##class Curry:
##    def __init__(self, fun, *args, **kwargs):
##        self.fun = fun
##        self.pending = args[:]
##        self.kwargs = kwargs.copy()
##
##    def __call__(self, *args, **kwargs):
##        if kwargs and self.kwargs:
##            kw = self.kwargs.copy()
##            kw.update(kwargs)
##        else:
##            kw = kwargs or self.kwargs
##        return self.fun(*(self.pending + args), **kw)
##

def getPythonVersion():
    return sys.version_info


PYTHON_VERSION = getPythonVersion()

if PYTHON_VERSION.major == 3:
    from io import BytesIO as StringIO
else:
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO


##
##def createStringBuffer(*args):
##    """Create a string with file-like behaviour (StringIO on Python 2.x).
##    """
##    return StringIO(*args)
##
##
##def binExtractor(fname, offset, length):
##    """Extract a junk of data from a file.
##    """
##    fp = open(fname)
##    fp.seek(offset)
##    data = fp.read(length)
##    return data
##


##
##CYG_PREFIX = "/cygdrive/"
##
##
##def cygpathToWin(path):
##    if path.startswith(CYG_PREFIX):
##        path = path[len(CYG_PREFIX):]
##        driveLetter = "{0}:\\".format(path[0])
##        path = path[2:].replace("/", "\\")
##        path = "{0}{1}".format(driveLetter, path)
##    return path
##
##
##class StructureWithEnums(ctypes.Structure):
##    """Add missing enum feature to ctypes Structures.
##    """
##    _map = {}
##
##    def __getattribute__(self, name):
##        _map = ctypes.Structure.__getattribute__(self, '_map')
##        value = ctypes.Structure.__getattribute__(self, name)
##        if name in _map:
##            EnumClass = _map[name]
##            if isinstance(value, ctypes.Array):
##                return [EnumClass(x) for x in value]
##            else:
##                return EnumClass(value)
##        else:
##            return value
##
##    def __str__(self):
##        result = []
##        result.append("struct {0} {{".format(self.__class__.__name__))
##        for field in self._fields_:
##            attr, attrType = field
##            if attr in self._map:
##                attrType = self._map[attr]
##            value = getattr(self, attr)
##            result.append("    {0} [{1}] = {2!r};".format(
##                attr, attrType.__name__, value))
##        result.append("};")
##        return '\n'.join(result)
##
##    __repr__ = __str__
##
##
##class CommandError(Exception):
##    pass
##
##
##def runCommand(cmd):
##    proc = subprocess.Popen(
##        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
##    result = proc.communicate()
##    proc.wait()
##    if proc.returncode:
##        raise CommandError("{0}".format(result[1]))
##    return result[0]
##
##
##class SingletonBase(object):
##    _lock = threading.Lock()
##
##    def __new__(cls, *args, **kws):
##        # Double-Checked Locking
##        if not hasattr(cls, '_instance'):
##            try:
##                cls._lock.acquire()
##                if not hasattr(cls, '_instance'):
##                    cls._instance = super(SingletonBase, cls).__new__(cls)
##            finally:
##                cls._lock.release()
##        return cls._instance
##


#  class RepresentationMixIn(object):
#
#      def __repr__(self):
#          keys = [k for k in self.__dict__ if not (k.startswith('__') and k.endswith('__'))]
#          result = []
#          result.append("{0!s} {{".format(self.__class__.__name__))
#          for key in keys:
#              value = getattr(self, key)
#              if isinstance(value, (int, )):
#                  line = "    {0!s} = 0x{1:X}".format(key, value)
#              elif isinstance(value, (float, types.NoneType)):
#                  line = "    {0!s} = {1!s}".format(key, value)
#              elif isinstance(value, array):
#                  line = "    {0!s} = {1!s}".format(key, hexDump(value))
#              else:
#                  line = "    {0!s} = '{1!s}'".format(key, value)
#              result.append(line)
#          result.append("}")
#          return '\n'.join(result)
#

##
##def memoryMap(filename, writeable=False):
##    size = os.path.getsize(filename)
##    fd = os.open(filename, os.O_RDWR if writeable else os.O_RDONLY)
##    return mmap.mmap(
##        fd, size, access=mmap.ACCESS_WRITE if writeable else mmap.ACCESS_READ)
##
##
##if sys.platform == "win32" and WINAPI:
##
##    # Code snippet taken from
##    # http://code.activestate.com/recipes/496767-set-process-priority-in-windows/
##    # Licenced under PSF.
##
##    def setpriority(pid=None, priority=1):
##        """ Set The Priority of a Windows Process.  Priority is a value
##            between 0-5 where 2 is normal priority.  Default sets the priority
##            of the current python process but can take any valid process ID."""
##        priorityclasses = [win32process.IDLE_PRIORITY_CLASS,
##                           win32process.BELOW_NORMAL_PRIORITY_CLASS,
##                           win32process.NORMAL_PRIORITY_CLASS,
##                           win32process.ABOVE_NORMAL_PRIORITY_CLASS,
##                           win32process.HIGH_PRIORITY_CLASS,
##                           win32process.REALTIME_PRIORITY_CLASS]
##        if pid is None:
##            # pid = win32api.GetCurrentProcessId()
##            pid = ctypes.windll.kernel32.GetCurrentProcessId()
##        handle = ctypes.windll.kernel32.OpenProcess(
##            win32con.PROCESS_ALL_ACCESS, True, pid)
##        win32process.SetPriorityClass(handle, priorityclasses[priority])
##        # Chris
##        win32process.SetProcessPriorityBoost(handle, False)
##        ct = ctypes.windll.kernel32.GetCurrentThread()
##        gle = win32api.GetLastError()
##        print("gct", win32api.FormatMessage(gle))
##        ctypes.windll.kernel32.SetThreadPriority(
##            ct, win32con.THREAD_BASE_PRIORITY_LOWRT)
##else:
##    def setpriority(pid=None, priority=1):
##        pass
##



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple
import enum

import construct

from construct import (
    Struct, Enum, Padding, Int8ul, GreedyBytes, Byte, Int16ul, Int32ul,
    BitStruct, BitsInteger, Flag, If, this, Int16ub, Int32ub, IfThenElse,
    Int16sl, Int32sl, Int16sb, Int32sb)


if construct.version < (2, 8):
    print("pyXCP requires at least construct 2.8")
    exit(1)


NumericType = (int, float)
MtaType = namedtuple("MtaType", "address ext")

class FrameSizeError(Exception):
    """
    A frame with an invalid size was received.
    """


class XcpResponseError(Exception):
    """
    Raise an `exception` from an XCP error packet.
    """

    def get_error_code(self):
        return self.args[0]


class XcpTimeoutError(Exception):
    """
    Timeout while waiting for a response occured.
    """


class XcpGetIdType(enum.IntEnum):
    ASCII_TEXT = 0
    FILENAME = 1
    FILE_AND_PATH = 2
    URL = 3
    FILE_TO_UPLOAD = 4


class Command(enum.IntEnum):

    # STD

    # Mandatory Commands
    CONNECT = 0xFF
    DISCONNECT = 0xFE
    GET_STATUS = 0xFD
    SYNCH = 0xFC

    # Optional Commands
    GET_COMM_MODE_INFO = 0xFB
    GET_ID = 0xFA
    SET_REQUEST = 0xF9
    GET_SEED = 0xF8
    UNLOCK = 0xF7
    SET_MTA = 0xF6
    UPLOAD = 0xF5
    SHORT_UPLOAD = 0xF4
    BUILD_CHECKSUM = 0xF3
    TRANSPORT_LAYER_CMD = 0xF2
    USER_CMD = 0xF1
    GET_VERSION = 0xC000

    # CAL

    # Mandatory Commands
    DOWNLOAD = 0xF0

    # Optional Commands
    DOWNLOAD_NEXT = 0xEF
    DOWNLOAD_MAX = 0xEE
    SHORT_DOWNLOAD = 0xED
    MODIFY_BITS = 0xEC

    # PAG

    # Mandatory Commands
    SET_CAL_PAGE = 0xEB
    GET_CAL_PAGE = 0xEA

    # Optional Commands
    GET_PAG_PROCESSOR_INFO = 0xE9
    GET_SEGMENT_INFO = 0xE8
    GET_PAGE_INFO = 0xE7
    SET_SEGMENT_MODE = 0xE6
    GET_SEGMENT_MODE = 0xE5
    COPY_CAL_PAGE = 0xE4

    # DAQ

    # Mandatory Commands
    CLEAR_DAQ_LIST = 0xE3
    SET_DAQ_PTR = 0xE2
    WRITE_DAQ = 0xE1
    WRITE_DAQ_MULTIPLE = 0xC7
    SET_DAQ_LIST_MODE = 0xE0
    GET_DAQ_LIST_MODE = 0xDF
    START_STOP_DAQ_LIST = 0xDE
    START_STOP_SYNCH = 0xDD

    # Optional Commands
    GET_DAQ_CLOCK = 0xDC
    READ_DAQ = 0xDB
    GET_DAQ_PROCESSOR_INFO = 0xDA
    GET_DAQ_RESOLUTION_INFO = 0xD9
    GET_DAQ_LIST_INFO = 0xD8
    GET_DAQ_EVENT_INFO = 0xD7
    DTO_CTR_PROPERTIES = 0xC5
    SET_DAQ_PACKED_MODE = 0xC001
    GET_DAQ_PACKED_MODE = 0xC002
    FREE_DAQ = 0xD6
    ALLOC_DAQ = 0xD5
    ALLOC_ODT = 0xD4
    ALLOC_ODT_ENTRY = 0xD3

    # PGM

    # Mandatory Commands
    PROGRAM_START = 0xD2
    PROGRAM_CLEAR = 0xD1
    PROGRAM = 0xD0
    PROGRAM_RESET = 0xCF

    # Optional Commands
    GET_PGM_PROCESSOR_INFO = 0xCE
    GET_SECTOR_INFO = 0xCD
    PROGRAM_PREPARE = 0xCC
    PROGRAM_FORMAT = 0xCB
    PROGRAM_NEXT = 0xCA
    PROGRAM_MAX = 0xC9
    PROGRAM_VERIFY = 0xC8

    TIME_CORRELATION_PROPERTIES = 0xC6


class CommandCategory(enum.IntEnum):
    """Values reflect resources (resource protection status / unlock).
    """
    STD     = 0
    CAL_PAG = 1
    DAQ     = 4
    STIM    = 8
    PGM     = 16



COMMAND_CATEGORIES = {  # Mainly needed to automatically UNLOCK.
    Command.CONNECT: CommandCategory.STD,
    Command.DISCONNECT: CommandCategory.STD,
    Command.GET_STATUS: CommandCategory.STD,
    Command.SYNCH: CommandCategory.STD,
    Command.GET_COMM_MODE_INFO: CommandCategory.STD,
    Command.GET_ID: CommandCategory.STD,
    Command.SET_REQUEST: CommandCategory.STD,
    Command.GET_SEED: CommandCategory.STD,
    Command.UNLOCK: CommandCategory.STD,
    Command.SET_MTA: CommandCategory.STD,
    Command.UPLOAD: CommandCategory.STD,
    Command.SHORT_UPLOAD: CommandCategory.STD,
    Command.BUILD_CHECKSUM: CommandCategory.STD,
    Command.TRANSPORT_LAYER_CMD: CommandCategory.STD,
    Command.USER_CMD: CommandCategory.STD,
    Command.GET_VERSION: CommandCategory.STD,

    Command.DOWNLOAD: CommandCategory.CAL_PAG,
    Command.DOWNLOAD_NEXT: CommandCategory.CAL_PAG,
    Command.DOWNLOAD_MAX: CommandCategory.CAL_PAG,
    Command.SHORT_DOWNLOAD: CommandCategory.CAL_PAG,
    Command.MODIFY_BITS: CommandCategory.CAL_PAG,

    Command.SET_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.GET_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.GET_PAG_PROCESSOR_INFO: CommandCategory.CAL_PAG,
    Command.GET_SEGMENT_INFO: CommandCategory.CAL_PAG,
    Command.GET_PAGE_INFO: CommandCategory.CAL_PAG,
    Command.SET_SEGMENT_MODE: CommandCategory.CAL_PAG,
    Command.GET_SEGMENT_MODE: CommandCategory.CAL_PAG,
    Command.COPY_CAL_PAGE: CommandCategory.CAL_PAG,
    Command.CLEAR_DAQ_LIST: CommandCategory.DAQ,
    Command.CLEAR_DAQ_LIST: CommandCategory.DAQ,
    Command.SET_DAQ_PTR: CommandCategory.DAQ,
    Command.WRITE_DAQ: CommandCategory.DAQ,
    Command.WRITE_DAQ_MULTIPLE: CommandCategory.DAQ,
    Command.SET_DAQ_LIST_MODE: CommandCategory.DAQ,
    Command.GET_DAQ_LIST_MODE: CommandCategory.DAQ,
    Command.START_STOP_DAQ_LIST: CommandCategory.DAQ,
    Command.START_STOP_SYNCH: CommandCategory.DAQ,
    Command.GET_DAQ_CLOCK: CommandCategory.DAQ,
    Command.READ_DAQ: CommandCategory.DAQ,
    Command.GET_DAQ_PROCESSOR_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_RESOLUTION_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_LIST_INFO: CommandCategory.DAQ,
    Command.GET_DAQ_EVENT_INFO: CommandCategory.DAQ,
    Command.DTO_CTR_PROPERTIES: CommandCategory.DAQ,
    Command.SET_DAQ_PACKED_MODE: CommandCategory.DAQ,
    Command.GET_DAQ_PACKED_MODE: CommandCategory.DAQ,
    Command.FREE_DAQ: CommandCategory.DAQ,
    Command.ALLOC_DAQ: CommandCategory.DAQ,
    Command.ALLOC_ODT: CommandCategory.DAQ,
    Command.ALLOC_ODT_ENTRY: CommandCategory.DAQ,

    Command.PROGRAM_START: CommandCategory.PGM,
    Command.PROGRAM_CLEAR: CommandCategory.PGM,
    Command.PROGRAM: CommandCategory.PGM,
    Command.PROGRAM_RESET: CommandCategory.PGM,
    Command.GET_PGM_PROCESSOR_INFO: CommandCategory.PGM,
    Command.GET_SECTOR_INFO: CommandCategory.PGM,
    Command.PROGRAM_PREPARE: CommandCategory.PGM,
    Command.PROGRAM_FORMAT: CommandCategory.PGM,
    Command.PROGRAM_NEXT: CommandCategory.PGM,
    Command.PROGRAM_MAX: CommandCategory.PGM,
    Command.PROGRAM_VERIFY: CommandCategory.PGM,

    # Well... ?
    # TIME_CORRELATION_PROPERTIES
}

XcpError = Enum(
    Int8ul,
    ERR_CMD_SYNCH=0x00,  # Command processor synchronization. S0

    ERR_CMD_BUSY=0x10,  # Command was not executed. S2
    ERR_DAQ_ACTIVE=0x11,  # Command rejected because DAQ is running. S2
    ERR_PGM_ACTIVE=0x12,  # Command rejected because PGM is running. S2

    ERR_CMD_UNKNOWN=0x20,  # Unknown command or not implemented optional
                           # command. S2
    ERR_CMD_SYNTAX=0x21,  # Command syntax invalid. S2
    ERR_OUT_OF_RANGE=0x22,  # Command syntax valid but command parameter(s)
                            # out of range. S2
    ERR_WRITE_PROTECTED=0x23,  # The memory location is write protected. S2
    ERR_ACCESS_DENIED=0x24,  # The memory location is not accessible. S2
    ERR_ACCESS_LOCKED=0x25,  # Access denied, Seed & Key is required. S2
    ERR_PAGE_NOT_VALID=0x26,  # Selected page not available. S2
    ERR_MODE_NOT_VALID=0x27,  # Selected page mode not available. S2
    ERR_SEGMENT_NOT_VALID=0x28,  # Selected segment not valid. S2
    ERR_SEQUENCE=0x29,  # Sequence error. S2
    ERR_DAQ_CONFIG=0x2A,  # DAQ configuration not valid. S2

    ERR_MEMORY_OVERFLOW=0x30,  # Memory overflow error. S2
    ERR_GENERIC=0x31,  # Generic error. S2
    ERR_VERIFY=0x32,  # The slave internal program verify routine detects an
                      # error. S3

    # NEW IN 1.1
    ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE=0x33,
    # Access to the requested resource is temporary not possible. S3

    ERR_TIMEOUT=0xff,   # Used by errorhandler; not an offical errorcode.
)

Response = Struct(
    "type" / Enum(
        Int8ul,
        OK=0xff,
        ERR=0xfe,
        EV=0xfd,
        SERV=0xfc,
    )
)

DAQ = Struct(
    "odt" / Byte,
    "daq" / Byte,
    "data" / GreedyBytes,
)

ResourceType = BitStruct(
    Padding(2),
    "dbg" / Flag,
    "pgm" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    Padding(1),
    "calpag" / Flag,
)

AddressGranularity = Enum(
    BitsInteger(2),
    BYTE=0b00,
    WORD=0b01,
    DWORD=0b10,
    RESERVED=0b11
)

ByteOrder = Enum(
    BitsInteger(1),
    INTEL=0,
    MOTOROLA=1
)

# byte-order dependent types
Int16u = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int16ul, Int16ub)
Int16s = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int16sl, Int16sb)
Int32u = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int32ul, Int32ub)
Int32s = IfThenElse(this._.byteOrder == ByteOrder.INTEL, Int32sl, Int32sb)

CommModeBasic = BitStruct(
    "optional" / Flag,  # The OPTIONAL flag indicates whether additional
                        # information on supported types of Communication mode
                        # is available. The master can get that additional
                        # information with GET_COMM_MODE_INFO
    "slaveBlockMode" / Flag,
    Padding(3),
    "addressGranularity" / AddressGranularity,
    "byteOrder" / ByteOrder
)

ConnectResponsePartial = Struct(
    "resource" / ResourceType,
    "commModeBasic" / CommModeBasic
)

ConnectResponse = Struct(
    "resource" / ResourceType,
    "commModeBasic" / CommModeBasic,
    "maxCto" / Int8ul,
    "maxDto" / Int16u,
    "protocolLayerVersion" / Int8ul,
    "transportLayerVersion" / Int8ul
)

GetVersionResponse = Struct(
    Padding(1),
    "protocolMajor" / Int8ul,
    "protocolMinor" / Int8ul,
    "transportMajor" / Int8ul,
    "transportMinor" / Int8ul,
)

SessionStatus = BitStruct(
    "resume" / Flag,
    "daqRunning" / Flag,
    Padding(2),
    "clearDaqRequest" / Flag,
    "storeDaqRequest" / Flag,
    Padding(1),
    "storeCalRequest" / Flag,
)

GetStatusResponse = Struct(
    "sessionStatus" / SessionStatus,
    "resourceProtectionStatus" / ResourceType,
    Padding(1),
    "sessionConfiguration" / Int16u,
)

CommModeOptional = BitStruct(
    Padding(6),
    "interleavedMode" / Flag,
    "masterBlockMode" / Flag,
)

GetCommModeInfoResponse = Struct(
    Padding(1),
    "commModeOptional" / CommModeOptional,
    Padding(1),
    "maxBs" / Int8ul,
    "minSt" / Int8ul,
    "queueSize" / Int8ul,
    "xcpDriverVersionNumber" / Int8ul,
)

GetIDResponse = Struct(
    "mode" / Int8ul,
    Padding(2),
    "length" / Int32u,
    "identification" / If(this.mode == 1, Byte[this.length])
)

GetSeedResponse = Struct(
    "length" / Int8ul,
    "seed" / If(this.length > 0, Byte[this.length])
)

SetRequestMode = BitStruct(
    Padding(4),
    "clearDaqReq" / Flag,
    "storeDaqReq" / Flag,
    Padding(1),
    "storeCalReq" / Flag,
)

BuildChecksumResponse = Struct(
    "checksumType" / Enum(
        Int8ul,
        XCP_NONE=0x00,
        XCP_ADD_11=0x01,
        XCP_ADD_12=0x02,
        XCP_ADD_14=0x03,
        XCP_ADD_22=0x04,
        XCP_ADD_24=0x05,
        XCP_ADD_44=0x06,
        XCP_CRC_16=0x07,
        XCP_CRC_16_CITT=0x08,
        XCP_CRC_32=0x09,
        XCP_USER_DEFINED=0xFF,
    ),
    Padding(2),
    "checksum" / Int32u,
)

SetCalPageMode = BitStruct(
    "all" / Flag,
    Padding(5),
    "xcp" / Flag,
    "ecu" / Flag,
)

GetPagProcessorInfoResponse = Struct(
    "maxSegments" / Int8ul,
    "pagProperties" / Int8ul,
)

GetSegmentInfoMode0Response = Struct(
    Padding(3),
    "basicInfo" / Int32u,
)

GetSegmentInfoMode1Response = Struct(
    "maxPages" / Int8ul,
    "addressExtension" / Int8ul,
    "maxMapping" / Int8ul,
    "compressionMethod" / Int8ul,
    "encryptionMethod" / Int8ul,
)

GetSegmentInfoMode2Response = Struct(
    Padding(3),
    "mappingInfo" / Int32u,
)

PageProperties = BitStruct(
    Padding(2),
    "xcpWriteAccessWithEcu" / Flag,
    "xcpWriteAccessWithoutEcu" / Flag,
    "xcpReadAccessWithEcu" / Flag,
    "xcpReadAccessWithoutEcu" / Flag,
    "ecuAccessWithXcp" / Flag,
    "ecuAccessWithoutXcp" / Flag,
)

DaqProperties = BitStruct(
    "overloadEvent" / Flag,
    "overloadMsb" / Flag,
    "pidOffSupported" / Flag,
    "timestampSupported" / Flag,
    "bitStimSupported" / Flag,
    "resumeSupported" / Flag,
    "prescalerSupported" / Flag,
    "daqConfigType" / Enum(
        BitsInteger(1),
        STATIC  = 0b0,
        DYNAMIC = 0b1
    ),
)

GetDaqProcessorInfoResponse = Struct(
    "daqProperties" / DaqProperties,
    "maxDaq" / Int16u,
    "maxEventChannel" / Int16u,
    "minDaq" / Int8ul,
    "daqKeyByte" / BitStruct(
        "Identification_Field" / Enum(
            BitsInteger(2),
            IDF_ABS_ODT_NUMBER=0b00,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_BYTE=0b01,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD=0b10,
            IDF_REL_ODT_NUMBER_ABS_DAQ_LIST_NUMBER_WORD_ALIGNED=0b11,
        ),
        "Address_Extension" / Enum(
            BitsInteger(2),
            AE_DIFFERENT_WITHIN_ODT=0b00,
            AE_SAME_FOR_ALL_ODT=0b01,
            _NOT_ALLOWED=0b10,
            AE_SAME_FOR_ALL_DAQ=0b11,
        ),
        "Optimisation_Type" / Enum(
            BitsInteger(4),
            OM_DEFAULT=0b0000,
            OM_ODT_TYPE_16=0b0001,
            OM_ODT_TYPE_32=0b0010,
            OM_ODT_TYPE_64=0b0011,
            OM_ODT_TYPE_ALIGNMENT=0b0100,
            OM_MAX_ENTRY_SIZE=0b0101,
        ),
    ),
)

DAQ_DIRECTION = Enum(
    BitsInteger(1),
    DAQ=0,
    STIM=1
)

PID_OFF = Enum(
    BitsInteger(1),
    DTO_WITH_ID_FIELD=0,
    DTO_WITHOUT_ID_FIELD=1
)

CurrentMode = BitStruct(
    "resume" / Flag,
    "running" / Flag,
    "pid_off" / PID_OFF,
    "timestamp" / Flag,
    Padding(2),
    "direction" / DAQ_DIRECTION,
    "selected" / Flag,
)

GetDaqListModeResponse = Struct(
    "currentMode" / CurrentMode,
    Padding(2),
    "currentEventChannel" / Int16u,
    "currentPrescaler" / Int8ul,
    "currentPriority" / Int8ul,
)

SetDaqListMode = BitStruct(
    Padding(2),
    "pid_off" / PID_OFF,
    "enable_timestamp" / Flag,
    Padding(2),
    "direction" / DAQ_DIRECTION,
    Padding(1)
)

DaqElement = Struct(
    "bitOffset" / Int8ul,
    "size" / Int8ul,
    "address" / Int32u,
    "addressExt" / Int8ul,
    Padding(1)
)

GetDaqClockResponse = Struct(
    Padding(3),
    "timestamp" / Int32u,
)

DaqPackedMode = Enum(
    Int8ul,
    NONE=0,
    ELEMENT_GROUPED=1,
    EVENT_GROUPED=2
)

GetDaqPackedModeResponse = Struct(
    Padding(1),
    "daqPackedMode" / DaqPackedMode,
    "dpmTimestampMode" / If(
        (this.daqPackedMode == "ELEMENT_GROUPED")
        | (this.daqPackedMode == "EVENT_GROUPED"),
        Int8ul
    ),
    "dpmSampleCount" / If(
        (this.daqPackedMode == "ELEMENT_GROUPED")
        | (this.daqPackedMode == "EVENT_GROUPED"),
        Int16u
    )
)

ReadDaqResponse = Struct(
    "bitOffset" / Int8ul,
    "sizeofDaqElement" / Int8ul,
    "adressExtension" / Int8ul,
    "address" / Int32u,
)

DaqTimestampUnit = Enum(
    BitsInteger(4),
    DAQ_TIMESTAMP_UNIT_1NS=0b0000,
    DAQ_TIMESTAMP_UNIT_10NS=0b0001,
    DAQ_TIMESTAMP_UNIT_100NS=0b0010,
    DAQ_TIMESTAMP_UNIT_1US=0b0011,
    DAQ_TIMESTAMP_UNIT_10US=0b0100,
    DAQ_TIMESTAMP_UNIT_100US=0b0101,
    DAQ_TIMESTAMP_UNIT_1MS=0b0110,
    DAQ_TIMESTAMP_UNIT_10MS=0b0111,
    DAQ_TIMESTAMP_UNIT_100MS=0b1000,
    DAQ_TIMESTAMP_UNIT_1S=0b1001,
    DAQ_TIMESTAMP_UNIT_1PS=0b1010,
    DAQ_TIMESTAMP_UNIT_10PS=0b1011,
    DAQ_TIMESTAMP_UNIT_100PS=0b1100
)

GetDaqResolutionInfoResponse = Struct(
    "granularityOdtEntrySizeDaq" / Int8ul,
    "maxOdtEntrySizeDaq" / Int8ul,
    "granularityOdtEntrySizeStim" / Int8ul,
    "maxOdtEntrySizeStim" / Int8ul,
    "timestampMode" / BitStruct(  # Int8ul,
        "unit" / DaqTimestampUnit,
        "fixed" / Flag,
        "size" / Enum(
            BitsInteger(3),
            NO_TIME_STAMP=0b000,
            S1=0b001,
            S2=0b010,
            NOT_ALLOWED=0b011,
            S4=0b100,
        ),
    ),
    "timestampTicks" / Int16u,
)

DaqListProperties = BitStruct(
    Padding(3),
    "packed" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    "eventFixed" / Flag,
    "predefined" / Flag,
)

GetDaqListInfoResponse = Struct(
    "daqListProperties" / DaqListProperties,
    "maxOdt" / Int8ul,
    "maxOdtEntries" / Int8ul,
    "fixedEvent" / Int16u,
)

StartStopDaqListResponse = Struct(
    "firstPid" / Int8ul
)

DaqEventProperties = BitStruct(
    "consistency" / Enum(
        BitsInteger(2),
        CONSISTENCY_ODT=0b00,
        CONSISTENCY_DAQ=0b01,
        CONSISTENCY_EVENTCHANNEL=0b10,
        CONSISTENCY_NONE=0b11,
    ),
    Padding(1),
    "packed" / Flag,
    "stim" / Flag,
    "daq" / Flag,
    Padding(2)
)

GetEventChannelInfoResponse = Struct(
    "daqEventProperties" / DaqEventProperties,
    "maxDaqList" / Int8ul,
    "eventChannelNameLength" / Int8ul,
    "eventChannelTimeCycle" / Int8ul,
    "eventChannelTimeUnit" / Enum(Int8ul,
        EVENT_CHANNEL_TIME_UNIT_1NS    = 0,
        EVENT_CHANNEL_TIME_UNIT_10NS   = 1,
        EVENT_CHANNEL_TIME_UNIT_100NS  = 2,
        EVENT_CHANNEL_TIME_UNIT_1US    = 3,
        EVENT_CHANNEL_TIME_UNIT_10US   = 4,
        EVENT_CHANNEL_TIME_UNIT_100US  = 5,
        EVENT_CHANNEL_TIME_UNIT_1MS    = 6,
        EVENT_CHANNEL_TIME_UNIT_10MS   = 7,
        EVENT_CHANNEL_TIME_UNIT_100MS  = 8,
        EVENT_CHANNEL_TIME_UNIT_1S     = 9,
        EVENT_CHANNEL_TIME_UNIT_1PS    = 10,
        EVENT_CHANNEL_TIME_UNIT_10PS   = 11,
        EVENT_CHANNEL_TIME_UNIT_100PS  = 12,
    ),
    "eventChannelPriority" / Int8ul,
)

DtoCtrProperties = BitStruct(
    "evtCtrPresent" / Flag,
    "stimCtrCpyPresent" / Flag,
    "stimModePresent" / Flag,
    "daqModePresent" / Flag,
    "relatedEventPresent" / Flag,
    "stimModeFixed" / Flag,
    "daqModeFixed" / Flag,
    "relatedEventFixed" / Flag
)

DtoCtrMode = BitStruct(
    Padding(6),
    "stimMode" / Flag,
    "daqMode" / Flag
)

DtoCtrPropertiesResponse = Struct(
    "properties" / DtoCtrProperties,
    "relatedEventChannel" / Int16u,
    "mode" / DtoCtrMode
)

CommModePgm = BitStruct(
    Padding(1),
    "slaveBlockMode" / Flag,
    Padding(4),
    "interleavedMode" / Flag,
    "masterBlockMode" / Flag,
)

ProgramStartResponse = Struct(
    Padding(1),
    "commModePgm" / CommModePgm,
    "maxCtoPgm" / Int8ul,
    "maxBsPgm" / Int8ul,
    "minStPgm" / Int8ul,
    "queueSizePgm" / Int8ul,
)

PgmProperties = BitStruct(
    "nonSeqPgmRequired" / Flag,
    "nonSeqPgmSupported" / Flag,
    "encryptionRequired" / Flag,
    "encryptionSupported" / Flag,
    "compressionRequired" / Flag,
    "compressionSupported" / Flag,
    "functionalMode" / Flag,
    "absoluteMode" / Flag
)

GetPgmProcessorInfoResponse = Struct(
    "pgmProperties" / PgmProperties,
    "maxSector" / Int8ul
)

GetSectorInfoResponseMode01 = Struct(
    "clearSequenceNumber" / Int8ul,
    "programSequenceNumber" / Int8ul,
    "programmingMethod" / Int8ul,
    "sectorInfo" / Int32u,
)

GetSectorInfoResponseMode2 = Struct(
    "sectorNameLength" / Int8ul
)

TimeCorrelationPropertiesResponse = Struct(
    "slaveConfig" / Int8ul,
    "observableClocks" / Int8ul,
    "syncState" / Int8ul,
    "clockInfo" / Int8ul,
    Padding(1),
    "clusterId" / Int16u
)

DaqPtr = namedtuple("DaqPtr", "daqListNumber odtNumber odtEntryNumber")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Very basic hello-world example.
"""

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""



from pprint import pprint

from pyxcp.cmdline import ArgumentParser

ap = ArgumentParser(description = "pyXCP hello world.")
with ap.run() as x:
    x.connect()
    if x.slaveProperties.optionalCommMode:
        x.getCommModeInfo()
    gid = x.getId(0x1)
    result = x.fetch(gid.length)
    x.disconnect()
    print("\nSlave properties:")
    print("=================")
    print("ID: '{}'".format(result.decode("utf8")))
    pprint(x.slaveProperties)



#!/bin/env python

import os
from setuptools import find_packages, setup

with open(os.path.join('pyxcp', 'version.py'), 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.split('=')[-1].strip().strip('"')
            break

with open("README.md", "r") as fh:
    long_description = fh.read()

install_reqs = [
    'pyusb', 'construct >= 2.9.0', 'mako', 'pyserial', 'toml', ]
setup_reqs = [
    'pyusb', 'construct >= 2.9.0', 'mako', 'pyserial', 'toml', ]

setup(
    name='pyxcp',
    version=version,
    provides=['pyxcp'],
    description="Universal Calibration Protocol for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Christoph Schueler',
    author_email='cpu12.gems@googlemail.com',
    url='https://github.com/christoph2/pyxcp',
    packages=find_packages(),

    include_package_data=True,
    install_requires=install_reqs,
    setup_requires=setup_reqs,
    extras_require={
       "docs": [
           'sphinxcontrib-napoleon'
       ],
        "develop": [
            "bumpversion"
       ]
    } ,
    package_dir={'tests': 'pyxcp/tests'},
    tests_require=["pytest", "pytest-runner"],
    test_suite="pyxcp.tests",
    license='GPLv2',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Types and structures to support error-handling as specified by XCP.
"""

__copyright__="""
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple
import enum

from pyxcp.types import Command, XcpError

Handler = namedtuple("Handler", "preAction  action")

TIMEOUT = 255

class PreAction(enum.IntEnum):
    """Pre-action to be taken, s. XCP spec.
    """
    NONE            = 0
    WAIT_T7         = 1
    SYNCH           = 2
    GET_SEED_UNLOCK = 3
    SET_MTA         = 4
    SET_DAQ_PTR     = 6
    START_STOP_X    = 7
    REINIT_DAQ      = 8
    DISPLAY_ERROR   = 9
    DOWNLOAD        = 10
    PROGRAM         = 11
    UPLOAD          = 12
    UNLOCK_SLAVE    = 13


class Action(enum.IntEnum):
    """Action to be taken, s. XCP spec.
    """
    NONE                = 0
    DISPLAY_ERROR       = 1
    RETRY_SYNTAX        = 2
    RETRY_PARAM         = 3
    USE_A2L             = 4     # Please refer to your A2L database.
    USE_ALTERATIVE      = 5
    REPEAT              = 6
    REPEAT_2_TIMES      = 7
    REPEAT_INF_TIMES    = 8
    RESTART_SESSION     = 9
    TERMINATE_SESSION   = 10
    SKIP                = 11
    NEW_FLASH_WARE      = 12

class Timeout(enum.IntEnum):
    """Various timeouts, s. XCP spec.
    """
    T1 = 0
    T2 = 1
    T3 = 2
    T4 = 3
    T5 = 4
    T6 = 5
    T7 = 6

class Severity(enum.IntEnum):
    """Severity of error.
    ---
    S0  = Information
    S1  = Warning / Request
    S2  = Resolvable Error
    S3  = Fatal Error
    """
    S0 = 0
    S1 = 1
    S2 = 2
    S3 = 3


ERROR_TABLE = {
    XcpError.ERR_CMD_SYNCH:         ("Command processor synchronization.",  Severity.S0),
    XcpError.ERR_CMD_BUSY:          ("Command was not executed.",  Severity.S2),
    XcpError.ERR_DAQ_ACTIVE:        ("Command rejected because DAQ is running.",  Severity.S2),
    XcpError.ERR_PGM_ACTIVE:        ("Command rejected because PGM is running.",  Severity.S2),
    XcpError.ERR_CMD_UNKNOWN:       ("Unknown command or not implemented optional command.",  Severity.S2),
    XcpError.ERR_CMD_SYNTAX:        ("Command syntax invalid",  Severity.S2),
    XcpError.ERR_OUT_OF_RANGE:      ("Command syntax valid but command parameter(s) out of range.",  Severity.S2),
    XcpError.ERR_WRITE_PROTECTED:   ("The memory location is write protected.",  Severity.S2),
    XcpError.ERR_ACCESS_DENIED:     ("The memory location is not accessible.",  Severity.S2),
    XcpError.ERR_ACCESS_LOCKED:     ("Access denied, Seed & Key is required", Severity.S2),
    XcpError.ERR_PAGE_NOT_VALID:    ("Selected page not available",  Severity.S2),
    XcpError.ERR_MODE_NOT_VALID:    ("Selected page mode not available",  Severity.S2),
    XcpError.ERR_SEGMENT_NOT_VALID: ("Selected segment not valid",  Severity.S2),
    XcpError.ERR_SEQUENCE:          ("Sequence error",  Severity.S2),
    XcpError.ERR_DAQ_CONFIG:        ("DAQ configuration not valid",  Severity.S2),
    XcpError.ERR_MEMORY_OVERFLOW:   ("Memory overflow error",  Severity.S2),
    XcpError.ERR_GENERIC:           ("Generic error.",  Severity.S2),
    XcpError.ERR_VERIFY:            ("The slave internal program verify routine detects an error.",  Severity.S3),
    XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: \
                                    ("Access to the requested resource is temporary not possible", Severity.S2),
}


ERROR_MATRIX = {
    Command.CONNECT         : { XcpError.ERR_TIMEOUT: ((PreAction.NONE, ), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),

                              },
    # CONNECT(USER_DEFINED)  timeout t6  wait t7  repeat  8 times

    Command.DISCONNECT      : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                              },
    Command.GET_STATUS      : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                },
    Command.SYNCH           : { XcpError.ERR_TIMEOUT: ((PreAction.NONE, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_SYNCH: ((PreAction.NONE), Action.SKIP),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.RESTART_SESSION),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                },
    Command.GET_COMM_MODE_INFO : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                   XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                   XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                   XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                 },
    Command.GET_ID            : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                  XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                  XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                  XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                },
    Command.SET_REQUEST     : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                },
    Command.GET_SEED       : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.UNLOCK         : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_ACCESS_LOCKED: ((PreAction.NONE), Action.RESTART_SESSION),
                               XcpError.ERR_SEQUENCE: ((PreAction.GET_SEED_UNLOCK), Action.REPEAT_2_TIMES),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.SET_MTA        : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.UPLOAD         : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.SHORT_UPLOAD   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.BUILD_CHECKSUM : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                               XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                               XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                               XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                               XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                               XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                               XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                             },
    Command.TRANSPORT_LAYER_CMD : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                    XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                    XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                    XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                    XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                    XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                  },
    Command.USER_CMD        : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.DOWNLOAD        : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.DOWNLOAD_NEXT   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.DOWNLOAD), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.SET_MTA), Action.USE_ALTERATIVE),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_SEQUENCE: ((PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.DOWNLOAD_MAX    : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.SET_MTA), Action.USE_ALTERATIVE),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                },
    Command.SHORT_DOWNLOAD  : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.MODIFY_BITS     : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.UPLOAD, PreAction.DOWNLOAD), Action.USE_ALTERATIVE),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.SET_CAL_PAGE    : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_CAL_PAGE    : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_PAG_PROCESSOR_INFO  : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                        XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                        XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                        XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                        XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                        XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                        XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                      },
    Command.GET_SEGMENT_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                              },
    Command.GET_PAGE_INFO   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                              },
    Command.SET_SEGMENT_MODE: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_SEGMENT_MODE: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.COPY_CAL_PAGE   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PAGE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.SET_DAQ_PTR     : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_DAQ_ACTIVE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.WRITE_DAQ       : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_DAQ_PTR), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.SET_DAQ_LIST_MODE: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                 XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                 XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                 XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                               },
    Command.START_STOP_DAQ_LIST: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                   XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                   XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                   XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                   XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                   XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                   XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                   XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                   XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                 },
    Command.START_STOP_SYNCH: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.CLEAR_DAQ_LIST  : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_DAQ_LIST_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                 XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                 XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                 XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                               },
    Command.WRITE_DAQ_MULTIPLE: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_DAQ_PTR), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                  XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                  XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                  XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                  XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                  XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_WRITE_PROTECTED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                  XcpError.ERR_DAQ_CONFIG: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                  XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                                },
    Command.READ_DAQ        : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_DAQ_PTR), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_DAQ_CLOCK   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                              },
    Command.GET_DAQ_PROCESSOR_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                      XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                      XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                      XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                      XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                      XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                      XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                    },
    Command.GET_DAQ_RESOLUTION_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                       XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                       XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                       XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                       XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                       XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                       XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                     },
    Command.GET_DAQ_LIST_MODE: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                 XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                 XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                 XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                 XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                               },
    Command.GET_DAQ_EVENT_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                  XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                  XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                  XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                  XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                  XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                  XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                },
    Command.FREE_DAQ        : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.ALLOC_DAQ       : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.ALLOC_ODT       : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.ALLOC_ODT_ENTRY : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.REINIT_DAQ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.REINIT_DAQ), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },

    Command.PROGRAM_START   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_DAQ_ACTIVE: ((PreAction.START_STOP_X), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_CLEAR   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM         : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_RESET   : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_PGM_ACTIVE: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.GET_PGM_PROCESSOR_INFO: { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                      XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                      XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                      XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                      XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                      XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                                    },
    Command.GET_SECTOR_INFO : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_A2L),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MODE_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_SEGMENT_NOT_VALID: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.NONE, Action.SKIP),
                              },
    Command.PROGRAM_PREPARE : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_FORMAT  : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_NEXT    : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.PROGRAM), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_DENIED: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_MAX     : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, PreAction.SET_MTA), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.USE_ALTERATIVE),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_MEMORY_OVERFLOW: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
    Command.PROGRAM_VERIFY  : { XcpError.ERR_TIMEOUT: ((PreAction.SYNCH, ), Action.REPEAT_2_TIMES),
                                XcpError.ERR_CMD_BUSY: ((PreAction.WAIT_T7), Action.REPEAT_INF_TIMES),
                                XcpError.ERR_CMD_UNKNOWN: ((PreAction.NONE), Action.DISPLAY_ERROR),
                                XcpError.ERR_CMD_SYNTAX: ((PreAction.NONE), Action.RETRY_SYNTAX),
                                XcpError.ERR_OUT_OF_RANGE: ((PreAction.NONE), Action.RETRY_PARAM),
                                XcpError.ERR_ACCESS_LOCKED: ((PreAction.UNLOCK_SLAVE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_SEQUENCE: ((PreAction.NONE), Action.REPEAT_2_TIMES),
                                XcpError.ERR_GENERIC: ((PreAction.NONE), Action.RESTART_SESSION),
                                XcpError.ERR_VERIFY: ((PreAction.NONE), Action.NEW_FLASH_WARE),
                                XcpError.ERR_RESOURCE_TEMPORARY_NOT_ACCESSIBLE: (PreAction.DISPLAY_ERROR, Action.REPEAT),
                              },
}

ALTERNATIVES = {
    Command.SHORT_UPLOAD:   (),
    Command.DOWNLOAD_NEXT:  (),
    Command.DOWNLOAD_MAX:   (),
    Command.SHORT_DOWNLOAD: (),
    Command.MODIFY_BITS:    (),
    Command.PROGRAM_NEXT:   (),
    Command.PROGRAM_MAX:    (),
}




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2019 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import selectors
import socket
import struct
from time import perf_counter, time

from pyxcp.transport.base import BaseTransport
import pyxcp.types as types

DEFAULT_XCP_PORT = 5555


class Eth(BaseTransport):
    """
    """

    PARAMETER_MAP = {
        #                  Type    Req'd   Default
        "HOST":           (str,    False,  "localhost"),
        "PORT":           (int,    False,  5555),
        "PROTOCOL":       (str,    False,  "TCP"),
        "IPV6":           (bool,   False,  False),
        "TCP_NODELAY":    (bool,   False,  False),
    }

    MAX_DATAGRAM_SIZE = 512
    HEADER = struct.Struct("<HH")
    HEADER_SIZE = HEADER.size

    def __init__(self, config=None):
        super(Eth, self).__init__(config)
        self.loadConfig(config)
        self.host = self.config.get("HOST")
        self.port = self.config.get("PORT")
        self.protocol = self.config.get("PROTOCOL")
        self.ipv6 = self.config.get("IPV6")
        self.use_tcp_no_delay = self.config.get("TCP_NODELAY")

        if self.ipv6 and not socket.has_ipv6:
            raise RuntimeError("IPv6 not supported by your platform.")
        else:
           address_family = socket.AF_INET6 if self.ipv6 else socket.AF_INET
        proto = socket.SOCK_STREAM if self.protocol == 'TCP' else socket.SOCK_DGRAM
        if self.host.lower() == "localhost":
            self.host = "::1" if self.ipv6 else "localhost"
        addrinfo = socket.getaddrinfo(self.host, self.port, address_family, proto)
        (self.address_family, self.socktype, self.proto, self.canonname, self.sockaddr) = addrinfo[0]
        self.status = 0
        self.sock = socket.socket(
            self.address_family,
            self.socktype,
            self.proto
        )
        self.selector = selectors.DefaultSelector()
        self.selector.register(self.sock, selectors.EVENT_READ)
        self.use_tcp = (self.protocol == 'TCP')
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if self.use_tcp and self.use_tcp_no_delay:
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "SO_REUSEPORT"):
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.sock.settimeout(0.5)

    def connect(self):
        if self.status == 0:
            self.sock.connect(self.sockaddr)
            self.startListener()
            self.status = 1  # connected

    def listen(self):
        HEADER_UNPACK = self.HEADER.unpack
        HEADER_SIZE = self.HEADER_SIZE
        use_tcp = self.use_tcp
        processResponse = self.processResponse
        EVENT_READ = selectors.EVENT_READ

        close_event_set = self.closeEvent.isSet
        socket_fileno = self.sock.fileno
        select = self.selector.select

        if use_tcp:
            sock_recv = self.sock.recv
        else:
            sock_recv = self.sock.recvfrom

        while True:
            try:
                if close_event_set() or socket_fileno() == -1:
                    return
                sel = select(0.1)
                for _, events in sel:
                    if events & EVENT_READ:
                        recv_timestamp = time()
                        if use_tcp:

                            # first try to get the header in one go
                            # if we are lucky this will avoid creating a
                            # bytearray and extending it
                            header = sock_recv(HEADER_SIZE)
                            size = len(header)
                            if size != HEADER_SIZE:

                                start = perf_counter()

                                header = bytearray(header)

                                while len(header) != HEADER_SIZE:
                                    header.extend(
                                        sock_recv(HEADER_SIZE - len(header))
                                    )
                                    if perf_counter() - start > 2:
                                        raise types.XcpTimeoutError("Eth frame header read timed out.") from None

                            length, counter = HEADER_UNPACK(header)

                            try:
                                # first try to get the response in one go
                                # similar to the header
                                response = sock_recv(length)
                                size = len(response)

                                if size != length:

                                    start = perf_counter()

                                    response = bytearray(response)
                                    while len(response) != length:
                                        response.extend(
                                            sock_recv(length - len(response))
                                        )

                                        if perf_counter() - start > 2:
                                            raise types.XcpTimeoutError("Eth frame payload read timed out.") from None

                            except Exception as e:
                                self.logger.error(str(e))
                                continue
                        else:
                            try:
                                response, _ = sock_recv(
                                    Eth.MAX_DATAGRAM_SIZE
                                )
                                length, counter = HEADER_UNPACK(
                                    response[:HEADER_SIZE]
                                )
                                response = response[HEADER_SIZE:]

                                if len(response) != length:
                                    raise types.FrameSizeError("Size mismatch.")

                            except Exception as e:
                                self.logger.error(str(e))
                                continue

                        processResponse(response, length, counter, recv_timestamp)
            except Exception:
                self.status = 0  # disconnected
                break

    def send(self, frame):
        self.pre_send_timestamp = time()
        self.sock.send(frame)
        self.post_send_timestamp = time()

    def closeConnection(self):
        if not self.invalidSocket:
            # Seems to be problematic /w IPv6
            # if self.status == 1:
            #     self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()

    @property
    def invalidSocket(self):
        return not hasattr(self, 'sock') or self.sock.fileno() == -1
